<?php

namespace Drupal\commerce_wechat_pay\Plugin\Commerce\PaymentGateway;

use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OffsitePaymentGatewayBase;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsRefundsInterface;
use Drupal\commerce_price\Price;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\HttpFoundation\Request;
use EasyWeChat\Foundation\Application;

/**
 * Provides WeChat QR-Code Payment Mode 2 gateway.
 * @link https://pay.weixin.qq.com/wiki/doc/api/native.php?chapter=6_5
 *
 * @CommercePaymentGateway(
 *   id = "qrcode_payment_mode_2",
 *   label = "WeChat QR-code Payment Mode 2",
 *   display_label = "WeChat QR-code Payment Mode 2",
 *    forms = {
 *     "offsite-payment" = "Drupal\commerce_wechat_pay\PluginForm\QRCodePaymentForm",
 *   }
 * )
 */
class QRCodePaymentMode2 extends OffsitePaymentGatewayBase implements SupportsRefundsInterface{

  const KEY_URI_PREFIX = 'private://commerce_wechat_pay/';

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['appid'] = [
      '#type' => 'textfield',
      '#title' => $this->t('公众账号ID'),
      '#description' => $this->t('绑定支付的APPID（开户邮件中可查看）'),
      '#default_value' => $this->configuration['appid'],
      '#required' => TRUE,
    ];

    $form['mch_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('商户号'),
      '#description' => $this->t('开户邮件中可查看'),
      '#default_value' => $this->configuration['mch_id'],
      '#required' => TRUE,
    ];

    $form['key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('商户支付密钥'),
      '#description' => $this->t('参考开户邮件设置（必须配置，登录商户平台自行设置）, 设置地址：https://pay.weixin.qq.com/index.php/account/api_cert'),
      '#default_value' => $this->configuration['key'],
      '#required' => TRUE,
    ];

    $form['sub_appid'] = [
      '#type' => 'textfield',
      '#title' => $this->t('子商户公众账号ID'),
      '#description' => $this->t('绑定支付的APPID（开户邮件中可查看）'),
      '#default_value' => $this->configuration['sub_appid']
    ];

    $form['sub_mch_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('子商户号'),
      '#description' => $this->t('开户邮件中可查看'),
      '#default_value' => $this->configuration['sub_mch_id']
    ];

    $form['certpem'] = [
      '#type' => 'textarea',
      '#title' => $this->t('证书文件内容'),
      '#description' => $this->t('Please open your apiclient_cert.pem file, and copy/paste the content to this text area.'),
      '#default_value' => $this->configuration['certpem']
    ];

    $form['keypem'] = [
      '#type' => 'textarea',
      '#title' => $this->t('证书文件密钥'),
      '#description' => $this->t('Please open your apiclient_key.pem file, and copy/paste the content to this text area.'),
      '#default_value' => $this->configuration['keypem']
    ];

    $form['certpem_uri'] = [
      '#type' => 'hidden',
      '#default_value' => $this->configuration['certpem_uri']
    ];

    $form['keypem_uri'] = [
      '#type' => 'hidden',
      '#default_value' => $this->configuration['keypem_uri']
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {

    $values = $form_state->getValue($form['#parents']);
    if (!empty($values['certpem'])) {
      // Check the file directory for storing cert/key files
      $path = self::KEY_URI_PREFIX;
      $dir = file_prepare_directory($path, FILE_CREATE_DIRECTORY);
      if (!$dir) {
        $form_state->setError($form['certpem'], $this->t('Commerce WeChat Pay cannot find your private file system. Please make sure your site has private file system configured!'));
        return;
      }

      $new_uri = $values['appid'] . md5($values['certpem']);

      if (empty($this->configuration['certpem_uri'])){
      } else {
        file_unmanaged_delete(self::KEY_URI_PREFIX . $this->configuration['certpem_uri']);
      }
      // We regenerate pem file in case the files were missing during server migration
      $updated = file_unmanaged_save_data($values['certpem'], self::KEY_URI_PREFIX . $new_uri);
      if ($updated) {
        $values['certpem_uri'] = $new_uri;
      } else {
        $form_state->setError($form['certpem'], $this->t('Commerce WeChat Pay cannot save your "certpem" into a file. Please make sure your site has private file system configured!'));
      }
    }

    if (!empty($values['keypem'])) {
      // Check the file directory for storing cert/key files
      $path = self::KEY_URI_PREFIX;
      $dir = file_prepare_directory($path, FILE_CREATE_DIRECTORY);
      if (!$dir) {
        $form_state->setError($form['keypem'], $this->t('Commerce WeChat Pay cannot find your private file system. Please make sure your site has private file system configured!'));
        return;
      }

      $new_uri = $values['appid'] . md5($values['keypem']);

      if (empty($this->configuration['keypem_uri'])) {
      } else {
        file_unmanaged_delete(self::KEY_URI_PREFIX . $this->configuration['keypem_uri']);
      }
      // We regenerate pem file in case the files were missing during server migration
      $updated = file_unmanaged_save_data($values['keypem'], self::KEY_URI_PREFIX . $new_uri);
      if ($updated) {
        $values['keypem_uri'] = $new_uri;
      } else {
        $form_state->setError($form['keypem'], $this->t('Commerce WeChat Pay cannot save your "keypem" into a file. Please make sure your site has private file system configured!'));
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);
    if (!$form_state->getErrors()) {
      $values = $form_state->getValue($form['#parents']);
      $this->configuration['appid'] = $values['appid'];
      $this->configuration['mch_id'] = $values['mch_id'];
      $this->configuration['key'] = $values['key'];
      $this->configuration['sub_appid'] = $values['sub_appid'];
      $this->configuration['sub_mch_id'] = $values['sub_mch_id'];
      $this->configuration['certpem'] = $values['certpem'];
      $this->configuration['keypem'] = $values['keypem'];
      $this->configuration['certpem_uri'] = $values['appid'] . md5($values['certpem']);
      $this->configuration['keypem_uri'] = $values['appid'] . md5($values['keypem']);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function refundPayment(PaymentInterface $payment, Price $amount = NULL) {
    if (!in_array($payment->getState()->value, ['capture_completed', 'capture_partially_refunded'])) {
      throw new \InvalidArgumentException(t('Only payments in the "capture_completed" and "capture_partially_refunded" states can be refunded.'));
    }
    // If not specified, refund the entire amount.
    $amount = $amount ?: $payment->getAmount();
    // Validate the requested amount.
    $balance = $payment->getBalance();
    if ($amount->greaterThan($balance)) {
      throw new InvalidRequestException(sprintf("Can't refund more than %s.", $balance->__toString()));
    }

    /** @var \Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OffsitePaymentGatewayInterface $payment_gateway_plugin */
    $payment_gateway_plugin = $payment->getPaymentGateway()->getPlugin();
    $appid = $payment_gateway_plugin->getConfiguration()['appid'];
    $mch_id = $payment_gateway_plugin->getConfiguration()['mch_id'];
    $key = $payment_gateway_plugin->getConfiguration()['key'];
    $sub_appid = $payment_gateway_plugin->getConfiguration()['appid'];
    $sub_mch_id = $payment_gateway_plugin->getConfiguration()['mch_id'];

    $cert_path = drupal_realpath(self::KEY_URI_PREFIX . $payment_gateway_plugin->getConfiguration()['certpem_uri']);
    $key_path = drupal_realpath(self::KEY_URI_PREFIX . $payment_gateway_plugin->getConfiguration()['keypem_uri']);

    if (!$cert_path || !$key_path) {
      throw new \InvalidArgumentException(t('Could not load the apiclient_cert.pem or apiclient_key.pem files, which are required for WeChat Refund. Did you configure them?'));
    }
    $options = [
      // 前面的appid什么的也得保留哦
      'app_id' => $appid,
      // ...
      // payment
      'payment' => [
        'merchant_id'        => $mch_id,
        'key'                => $key,
        'cert_path'          => $cert_path, // XXX: 绝对路径！！！！
        'key_path'           => $key_path,      // XXX: 绝对路径！！！！
        //'notify_url'         => '默认的订单回调地址',       // 你也可以在下单时单独设置来想覆盖它
        // 'device_info'     => '013467007045764',
        // 'sub_app_id'      => '',
        // 'sub_merchant_id' => '',
        // ...
      ],
    ];
    $app = new Application($options);
    $wechat_pay = $app->payment;

    $result = $wechat_pay->refund($payment->getOrderId(), $payment->getOrderId() . date("zHis"), floatval($payment->getOrder()->getTotalPrice()->getNumber()) * 100, floatval($amount->getNumber()) * 100);

    if (!$result->return_code == 'SUCCESS' || !$result->result_code == 'SUCCESS'){
      // For any reason, we cannot get a preorder made by WeChat service
      throw new \InvalidRequestException(t('WeChat Service cannot approve this request: ') . $result->err_code_des);
    }

    // Perform the refund request here, throw an exception if it fails.
    // See \Drupal\commerce_payment\Exception for the available exceptions.
    $remote_id = $payment->getRemoteId();
    $number = $amount->getNumber();

    $old_refunded_amount = $payment->getRefundedAmount();
    $new_refunded_amount = $old_refunded_amount->add($amount);
    if ($new_refunded_amount->lessThan($payment->getAmount())) {
      $payment->state = 'capture_partially_refunded';
    }
    else {
      $payment->state = 'capture_refunded';
    }

    $payment->setRefundedAmount($new_refunded_amount);
    $payment->save();
  }

  /**
   * {@inheritdoc}
   */
  public function onNotify(Request $request) {

    $appid = $this->getConfiguration()['appid'];
    $mch_id = $this->getConfiguration()['mch_id'];
    $key = $this->getConfiguration()['key'];
    $options = [
      'app_id' => $appid,
      'payment' => [
        'merchant_id'        => $mch_id,
        'key'                => $key,
      ],
    ];
    $app = new Application($options);

    $response = $app->payment->handleNotify(function($notify, $successful) {

      if ($successful) {
        $this->createPayment($notify->toArray());

      } else { // When payment failed
      }

      return TRUE; // Respond WeChat request that we have finished processing this notification
    });

    return $response;
  }

  /**
   * Create a Commerce Payment from a WeChat request successful result
   * @param  array $result
   * @param  string $state
   * @param  OrderInterface $order
   * @param  string $remote_state
   */
  public function createPayment(array $result, $state = 'capture_completed', OrderInterface $order = NULL, $remote_state = NULL) {
    /** @var \Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OffsitePaymentGatewayInterface $payment_gateway_plugin */
    $payment_storage = $this->entityTypeManager->getStorage('commerce_payment');

    if (!$order) {
      $price = new Price(strval($result['total_fee'] / 100), $result['fee_type']);
    } else {
      $price = $order->getTotalPrice();
    }
    $payment = $payment_storage->create([
      'state' => $state,
      'amount' => $price,
      'payment_gateway' => $this->entityId,
      'order_id' => $result['out_trade_no']? $result['out_trade_no'] : $order->id(),
      'test' => $this->getMode() == 'test',
      'remote_id' => $result['transaction_id'],
      'remote_state' => $remote_state,
      'authorized' => REQUEST_TIME
    ]);
    $payment->save();
  }

}
